	;ROUT KEYB MANAGER.ASM      18F family

READ_KEYB

;The reading of the keyboard consists of two scans of the keyboard matrix
;separated by a debounce delay.

;If the outcome of both scans match, the value is passed to the interpreter
;which checks it against all possible valid readings. When found, it jumps into
;the routine servicing the pressed key/s. If no key or an invalid combination
;of keys was pressed, the interpreter resets the "active-key" flag and returns.

;Since a multiple-keys press is possible, this routine is appropriate for the
;use of qualifier keys.

;The debouncing delay is based on a counter decremented in the ad-hoc routine,
;DECR_COUNTERS, inside the main loop.

  BTFSS SCAN_1ST_DONE     ;1st matrix scan, done already?
  BRA SCAN_1ST_TIME       ;not yet - do 1st scan!

  TSTFSZ CNTR_DEB_KEYB    ;1st scan done already - debounce delay elapsed?
  RETURN                  ;debounce delay not elapsed yet - repeat checking
  
  BRA SCAN_2ND_TIME       ;debounce delay elapsed - do 2nd scan!
;------------------------------------------------------------------------------

SCAN_1ST_TIME
  CLRF KEYB_SCAN          ;clear register used to held outcome of 1st scan
  CLRF KEYB_ROWS          ;clear register to record active rows
  LOADREG KEYB_COLS,B'11110000'   ;prepare register to record active columns

  CALL SCAN_KEYB_MATRIX   ;requests first matrix scan
  
  MOVF KEYB_COLS,W        ;W gets "active columns" info
  IORWF KEYB_ROWS,W       ;W incorporates "active rows" info
  MOVWF KEYB_SCAN         ;active columns & rows info to KEYB_SCAN

  BSF SCAN_1ST_DONE       ;tell the manager "first matrix scan is done"

  LOADREG	CNTR_DEB_KEYB,COUNT_DEB_KEYB    ;start debounce counter
  RETURN
;------------------------------------------------------------------------------

SCAN_2ND_TIME
  CLRF KEYB_ROWS          ;clear register to record active rows
  LOADREG KEYB_COLS,B'11110000'   ;prepare register to record active columns

  CALL SCAN_KEYB_MATRIX   ;requests second matrix scan
  
  MOVF KEYB_COLS,W        ;W gets "active columns" info
  IORWF KEYB_ROWS,W       ;W incorporates "active rows" info

  SUBWF KEYB_SCAN,W
  ;susbstract from first scan, to check if both scans match
  
  BTFSC STATUS,Z          ;Z =0, scans don't match - invalid reading!
  RCALL KEYB_INTERPRETER  ;Z =1, scans do match - process reading!

READ_KEYB_EXIT            ;reading (and processing, if needed) done!
  BCF SCAN_1ST_DONE       ;enables the start of the next keayboard reading
  RETURN
;------------------------------------------------------------------------------

SCAN_KEYB_MATRIX

;Hardware setup:

;(b7:b4) COLS, always IN, with weak pull-ups enabled.
;(b3:b0) ROWS, all initially IN, sequentially set as OUT to output a 0 during
;the reading, reverting, immediately after, to the IN condition.
;A pressed key, closes a contact, bringing the respective column to 0.

;The routine scans the keyboard matrix by reading the four keys in each row,
;starting with row R3 down to R0.

;The reading of each row is:
;a) ANDed to KEYB_COLS
;b) checked against B'11110000'. If it difers, (at least one column was =0),
; the respective row bit in KEYB_ROWS is set.

READ_ROW_3                  ;informative label
  BCF KEYB_DIR,3         ;R3 OUT - outputs a 0
  CALL KEYB_READ_ROW      ;read cols with row 3 outputing a 0
  BTFSC ROW_ACTIVE        ;if one or more columns were active...
  BSF KEYB_ROWS,3         ;...set row 3 as active
  BCF ROW_ACTIVE          ;clear flag to read next row
  BSF KEYB_DIR,3         ;R3 IN again

READ_ROW_2                  ;informative label
  BCF KEYB_DIR,2         ;R2 OUT - outputs a 0
  CALL KEYB_READ_ROW      ;read cols with row 2 outputing a 0
  BTFSC ROW_ACTIVE        ;if one or more columns were active...
  BSF KEYB_ROWS,2         ;...set row 2 as active
  BCF ROW_ACTIVE          ;clear flag to read next row
  BSF KEYB_DIR,2         ;R2 IN again

READ_ROW_1                  ;informative label
  BCF KEYB_DIR,1         ;R1 OUT - outputs a 0
  CALL KEYB_READ_ROW      ;read cols with row 1 outputing a 0
  BTFSC ROW_ACTIVE        ;if one or more columns were active...
  BSF KEYB_ROWS,1         ;...set row 1 as active
  BCF ROW_ACTIVE          ;clear flag to read next row
  BSF KEYB_DIR,1         ;R1 IN again

READ_ROW_0                  ;informative label
  BCF KEYB_DIR,0         ;R0 OUT - outputs a 0
  RCALL KEYB_READ_ROW     ;read cols with row 0 outputing a 0
  BTFSC ROW_ACTIVE        ;if one or more columns were active...
  BSF KEYB_ROWS,0         ;...set row 0 as active
  BCF ROW_ACTIVE          ;clear flag to leave info in KEYB_ROWS, intact
  BSF KEYB_DIR,0         ;R0 IN again
  RETURN                  ;cols info in KEYB_COLS / rows info in KEYB_ROWS
;------------------------------------------------------------------------------

KEYB_READ_ROW
  MOVLW B'11110000'       ;W is now able to read the four columns
  ANDWF KEYB_PORT,W       ;any key pressed makes its column low
  ANDWF KEYB_COLS,F       ;the reading is ANDed with the columns register
  SUBLW B'11110000'       ;substract to see if at least one column was low
  BTFSS STATUS,Z          ;if Z =1, no column was low - leave!
  BSF ROW_ACTIVE          ;Z =0, tell the main routine "this row was active"
  RETURN
;------------------------------------------------------------------------------

KEYB_INTERPRETER            ;to enter here, KEYB_SCAN with keyboard reading

  SKIP_R_DIF_V KEYB_SCAN,B'01111000'
      BRA KEYB_SERV_C7_R3

  SKIP_R_DIF_V KEYB_SCAN,B'01110100'
      BRA KEYB_SERV_C7_R2

  SKIP_R_DIF_V KEYB_SCAN,B'01110010'
      BRA KEYB_SERV_C7_R1

  SKIP_R_DIF_V KEYB_SCAN,B'01110001'
      BRA KEYB_SERV_C7_R0
;------------------------------------------------------------------------------
  SKIP_R_DIF_V KEYB_SCAN,B'10111000'
      BRA KEYB_SERV_C6_R3

  SKIP_R_DIF_V KEYB_SCAN,B'10110100'
      BRA KEYB_SERV_C6_R2

  SKIP_R_DIF_V KEYB_SCAN,B'10110010'
      BRA KEYB_SERV_C6_R1

  SKIP_R_DIF_V KEYB_SCAN,B'10110001'
      BRA KEYB_SERV_C6_R0
;------------------------------------------------------------------------------
  SKIP_R_DIF_V KEYB_SCAN,B'11011000'
      BRA KEYB_SERV_C5_R3

  SKIP_R_DIF_V KEYB_SCAN,B'11010100'
      BRA KEYB_SERV_C5_R2

  SKIP_R_DIF_V KEYB_SCAN,B'11010010'
      BRA KEYB_SERV_C5_R1

  SKIP_R_DIF_V KEYB_SCAN,B'11010001'
      BRA KEYB_SERV_C5_R0
;------------------------------------------------------------------------------
  SKIP_R_DIF_V KEYB_SCAN,B'11101000'
      BRA KEYB_SERV_C4_R3

  SKIP_R_DIF_V KEYB_SCAN,B'11100100'
      BRA KEYB_SERV_C4_R2

  SKIP_R_DIF_V KEYB_SCAN,B'11100010'
      BRA KEYB_SERV_C4_R1

  SKIP_R_DIF_V KEYB_SCAN,B'11100001'
      BRA KEYB_SERV_C4_R0

;arriving herebelow means "no key or a not allowed combination of keys was
;pressed", thus no key is active; flag that and escape...!
      BCF KEY_ACTIVE
      RETURN
;------------------------------------------------------------------------------  
